
//  LauncherOSX
//
//  Created by Boris Schneiderman.
//  Copyright (c) 2014 Readium Foundation and/or its licensees. All rights reserved.
//  
//  Redistribution and use in source and binary forms, with or without modification, 
//  are permitted provided that the following conditions are met:
//  1. Redistributions of source code must retain the above copyright notice, this 
//  list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation and/or 
//  other materials provided with the distribution.
//  3. Neither the name of the organization nor the names of its contributors may be 
//  used to endorse or promote products derived from this software without specific 
//  prior written permission.
//  
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
//  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
//  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
//  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
//  OF THE POSSIBILITY OF SUCH DAMAGE.

/*
 * Renders reflowable content using CSS columns
 *
 * @class ReadiumSDK.Views.ReflowableView
 */

ReadiumSDK.Views.ReflowableView = function(options, reader){

    _.extend(this, Backbone.Events);

    var self = this;
    
    var _$viewport = options.$viewport;
    var _spine = options.spine;
    var _userStyles = options.userStyles;
    var _bookStyles = options.bookStyles;
    var _iframeLoader = options.iframeLoader;
    
    var _currentSpineItem;
    var _isWaitingFrameRender = false;    
    var _deferredPageRequest;
    var _fontSize = 110;
    var _fontColor = "#000000";
    var _fontFamily = "lantinghei";
    var _$contentFrame;
    var _navigationLogic;
    var _$el;
    var _$iframe;
    var _$epubHtml;
    
    var _$htmlBody;
    
    var _htmlBodyIsVerticalWritingMode;
    var _htmlBodyIsLTRDirection;
    var _htmlBodyIsLTRWritingMode;
    
    var _currentOpacity = -1;

    var _lastViewPortSize = {
        width: undefined,
        height: undefined
    };

    var _paginationInfo = {

        visibleColumnCount : 2,
        columnGap : 20,
        spreadCount : 0,
        currentSpreadIndex : 0,
        columnWidth : undefined,
        pageOffset : 0,
        columnCount: 0
    };
    
    var _origImageSize;
    
    var _startTouch = undefined;
    var _lastTouch = undefined;
    var _hold = false;
    
    var _startTouch2 = undefined;
    var _lastTouch2 = undefined;
    
    var ratio = 1;
    //var ratio = window.devicePixelRatio;
    
    var _highlights = [];
    var _bookmarks = [];

    var _selectionRange;
    
    this.render = function(){
//        console.debug("render start");
        var template = ReadiumSDK.Helpers.loadTemplate("reflowable_book_frame", {});

        _$el = $(template);
        _$viewport.append(_$el);

        // This fixes rendering issues with WebView (native apps), which clips content embedded in iframes unless GPU hardware acceleration is enabled for CSS rendering.
        _$el.css("transform", "translateZ(0)");
        
        _$contentFrame = $("#reflowable-content-frame", _$el);

        _$iframe = $("#epubContentIframe", _$el);

        _$iframe.css("left", "");
        _$iframe.css("right", "");
        _$iframe.css("position", "relative");
        //_$iframe.css(_spine.isLeftToRight() ? "left" : "right", "0px");
        _$iframe.css("overflow", "hidden");

        _navigationLogic = new ReadiumSDK.Views.CfiNavigationLogic(
                _$contentFrame, _$iframe,
                { rectangleBased: true, paginationInfo: _paginationInfo });


        // See ReaderView.handleViewportResize
        // var lazyResize = _.debounce(self.onViewportResize, 100);
        // $(window).on("resize.ReadiumSDK.reflowableView", _.bind(lazyResize, self));

//        console.debug("render end");
        return self;
    };

    function setFrameSizesToRectangle(rectangle) {
        _$contentFrame.css("left", rectangle.left + "px");
        _$contentFrame.css("top", rectangle.top + "px");
        _$contentFrame.css("right", rectangle.right + "px");
        _$contentFrame.css("bottom", rectangle.bottom + "px");

    }

    this.remove = function() {

        //$(window).off("resize.ReadiumSDK.reflowableView");
        _$el.remove();

    };

    this.isReflowable = function() {
        return true;
    };

    this.onViewportResize = function() {

        if(updateViewportSize()) {
            updatePagination();
        }
    };

    var _viewSettings = undefined;
    this.setViewSettings = function(settings) {
        
        _viewSettings = settings;

        _paginationInfo.columnGap = settings.columnGap * ratio;
        _fontSize = settings.fontSize;
        _fontColor = settings.fontColor;
        _fontFamily = settings.fontFamily;

        updateHtmlFontSize();
        updateColumnGap();
        
        updateViewportSize();
        updatePagination();
    };

    function loadSpineItem(spineItem) {

        if(_currentSpineItem != spineItem) {

            _paginationInfo.pageOffset = 0;
            _paginationInfo.currentSpreadIndex = 0;
            _currentSpineItem = spineItem;
            _isWaitingFrameRender = true;

            var src = _spine.package.resolveRelativeUrl(spineItem.href);
                        
            self.trigger(ReadiumSDK.Events.CONTENT_DOCUMENT_LOAD_START, _$iframe, spineItem);

            _$iframe.css("opacity", "0.01");
            
            _iframeLoader.loadIframe(_$iframe[0], src, onIFrameLoad, self, {spineItem : spineItem});
        }
    }

    function updateHtmlFontSize() {

        if(_$epubHtml) {
            _$epubHtml.css("font-size", (_fontSize * ratio) + "%");
            _$epubHtml.css("font-family", _fontFamily + ",stix");
            _$epubHtml.css("line-height", 1.65);
            _$epubHtml.css("word-break", "break-all");
            _$epubHtml.css("letter-spacing", "2px");
           // _$epubHtml.css("text-align", "justify");
          //  _$epubHtml.css("text-justify", "inter-ideograph");
            _$epubHtml.css("color", _fontColor);
        }
        $('p', _$epubHtml).each(function(){
                               $elem = $(this);
                               $elem.css("line-height", 1.65);
                               $elem.css("font-size", "1em");
                               $elem.css("letter-spacing", "2px");
                               var align = $elem.css("text-align");
                               if(!align||align=="left"||align=="justify"){
                                   $elem.css("text-align", "justify");
                                   $elem.css("text-justify", "inter-ideograph");
                               }
                               var indent = $elem.css("text-indent");
                               if (parseFloat(indent) > 0){
                                   $elem.css("text-indent", "2em");
                               }
                           });        
        $('h1', _$epubHtml).each(function(){
                                $elem = $(this);
                                $elem.css("line-height", 1.2);
                                $elem.css("font-size", "1.2em");
                                $elem.css("text-align", "justify");
                               $elem.css("letter-spacing", "2px");
                                $elem.css("font-weight", "bold");
                                $elem.css("text-justify", "inter-ideograph");
                            });
        $('h2', _$epubHtml).each(function(){
                                $elem = $(this);
                                $elem.css("line-height", 1.3);
                                $elem.css("font-size", "1.1em");
                               $elem.css("letter-spacing", "2px");
                                $elem.css("text-align", "justify");
                                $elem.css("font-weight", "bold");
                                $elem.css("text-justify", "inter-ideograph");
                            });
        $('h3', _$epubHtml).each(function(){
                                $elem = $(this);
                                $elem.css("line-height", 1.3);
                                $elem.css("font-size", "1.1em");
                               $elem.css("letter-spacing", "2px");
                                $elem.css("text-align", "justify");
                                $elem.css("font-weight", "bold");
                                $elem.css("text-justify", "inter-ideograph");
                            });
        $('h4', _$epubHtml).each(function(){
                                $elem = $(this);
                                $elem.css("line-height", 1.3);
                                $elem.css("font-size", "1.1em");
                               $elem.css("letter-spacing", "2px");
                                $elem.css("text-align", "justify");
                                $elem.css("font-weight", "bold");
                                $elem.css("text-justify", "inter-ideograph");
                            });
        $('h5', _$epubHtml).each(function(){
                                $elem = $(this);
                                $elem.css("line-height", 1.3);
                                $elem.css("font-size", "1.1em");
                                $elem.css("text-align", "justify");
                               $elem.css("letter-spacing", "2px");
                                $elem.css("font-weight", "bold");
                                $elem.css("text-justify", "inter-ideograph");
                            });
        $('h6', _$epubHtml).each(function(){
                                $elem = $(this);
                                $elem.css("line-height", 1.3);
                                $elem.css("font-size", "1.1em");
                                $elem.css("text-align", "justify");
                               $elem.css("letter-spacing", "2px");
                                $elem.css("font-weight", "bold");
                                $elem.css("text-justify", "inter-ideograph");
                            });
        
        adjustImageSize();
    }

    function updateColumnGap() {

        if(_$epubHtml) {
        
            _.each(['-webkit-', '-moz-', '-ms-', ''], function(prefix) {
                _$epubHtml.css(prefix + "column-gap", _paginationInfo.columnGap + "px");
            });
        }
    }
    
    function onIFrameLoad(success) {

        _isWaitingFrameRender = false;

        //while we where loading frame new request came
        if(_deferredPageRequest && _deferredPageRequest.spineItem != _currentSpineItem) {
            loadSpineItem(_deferredPageRequest.spineItem);
            return;
        }

        if(!success) {
            _$iframe.css("opacity", "1");
            _deferredPageRequest = undefined;
            return;
        }

        self.trigger(ReadiumSDK.Events.CONTENT_DOCUMENT_LOADED, _$iframe, _currentSpineItem);

        var epubContentDocument = _$iframe[0].contentDocument;
        _$epubHtml = $("html", epubContentDocument);
        _$htmlBody = $("body", _$epubHtml);
        
        _htmlBodyIsVerticalWritingMode = false;
        _htmlBodyIsLTRDirection = true;
        _htmlBodyIsLTRWritingMode = undefined;
        
        var win = epubContentDocument.defaultView || _$iframe[0].contentWindow;
        
        //Helpers.isIframeAlive
        var htmlBodyComputedStyle = win.getComputedStyle(_$htmlBody[0], null);
        if (htmlBodyComputedStyle)
        {
            _htmlBodyIsLTRDirection = htmlBodyComputedStyle.direction === "ltr";

            var writingMode = undefined;
            if (htmlBodyComputedStyle.getPropertyValue)
            {
                writingMode = htmlBodyComputedStyle.getPropertyValue("-webkit-writing-mode") || htmlBodyComputedStyle.getPropertyValue("-moz-writing-mode") || htmlBodyComputedStyle.getPropertyValue("-ms-writing-mode") || htmlBodyComputedStyle.getPropertyValue("-o-writing-mode") || htmlBodyComputedStyle.getPropertyValue("-epub-writing-mode") || htmlBodyComputedStyle.getPropertyValue("writing-mode");
            }
            else
            {
                writingMode = htmlBodyComputedStyle.webkitWritingMode || htmlBodyComputedStyle.mozWritingMode || htmlBodyComputedStyle.msWritingMode || htmlBodyComputedStyle.oWritingMode || htmlBodyComputedStyle.epubWritingMode || htmlBodyComputedStyle.writingMode;
            }

            if (writingMode)
            {
                _htmlBodyIsLTRWritingMode = writingMode.indexOf("-lr") >= 0; // || writingMode.indexOf("horizontal-") >= 0; we need explicit!
            
                if (writingMode.indexOf("vertical") >= 0 || writingMode.indexOf("tb-") >= 0 || writingMode.indexOf("bt-") >= 0)
                {
                    _htmlBodyIsVerticalWritingMode = true;
                }
            }
        }

        if (_htmlBodyIsLTRDirection)
        {
            if (_$htmlBody[0].getAttribute("dir") === "rtl" || _$epubHtml[0].getAttribute("dir") === "rtl")
            {
                _htmlBodyIsLTRDirection = false;
            }
        }
         
        // Some EPUBs may not have explicit RTL content direction (via CSS "direction" property or @dir attribute) despite having a RTL page progression direction. Readium consequently tweaks the HTML in order to restore the correct block flow in the browser renderer, resulting in the appropriate CSS columnisation (which is used to emulate pagination).
        if (!_spine.isLeftToRight() && _htmlBodyIsLTRDirection && !_htmlBodyIsVerticalWritingMode)
        {
            _$htmlBody[0].setAttribute("dir", "rtl");
            _htmlBodyIsLTRDirection = false;
            _htmlBodyIsLTRWritingMode = false;
         }
        
        _paginationInfo.isVerticalWritingMode = _htmlBodyIsVerticalWritingMode;
        
        hideBook();
        _$iframe.css("opacity", "1");
        
        updateViewportSize();
        _$epubHtml.css("height", _lastViewPortSize.height + "px");
        
        _$epubHtml.css("position", "relative");
        _$epubHtml.css("margin", "0");
        _$epubHtml.css("padding", "0");

        _.each(['-webkit-', '-moz-', '-ms-', ''], function(prefix) {
            _$epubHtml.css(prefix + "column-axis", (_htmlBodyIsVerticalWritingMode ? "vertical" : "horizontal"));
        });

        //
        // /////////
        // //Columns Debugging
        //
        // _.each(['-webkit-', '-moz-', '-ms-', ''], function(prefix) {
        //     _$epubHtml.css(prefix + "column-rule-color", "red");
        //     _$epubHtml.css(prefix + "column-rule-style", "dashed");
        //     _$epubHtml.css(prefix + "column-rule-width", "1px");
        // });
        // _$epubHtml.css("background-color", '#b0c4de');
        //
        // ////
        
        
        self.applyBookStyles();
        updateMedias();
        resizeImages();
        
        updateHtmlFontSize();
        updateColumnGap();

        self.applyStyles();
        
        win.focus();
    }

    this.applyStyles = function() {

        ReadiumSDK.Helpers.setStyles(_userStyles.getStyles(), _$el.parent());

        //because left, top, bottom, right setting ignores padding of parent container
        //we have to take it to account manually
        var elementMargins = ReadiumSDK.Helpers.Margins.fromElement(_$el);
        setFrameSizesToRectangle(elementMargins.padding);

        updateViewportSize();
        
        updatePagination();
    };

    this.applyBookStyles = function() {

        if(_$epubHtml) {
            ReadiumSDK.Helpers.setStyles(_bookStyles.getStyles(), _$epubHtml);
        }
    };

    function openDeferredElement() {

        if(!_deferredPageRequest) {
            return;
        }

        var deferredData = _deferredPageRequest;
        _deferredPageRequest = undefined;
        self.openPage(deferredData);

    }

    this.openPage = function(pageRequest) {

        if(_isWaitingFrameRender) {
            _deferredPageRequest = pageRequest;
            return;
        }

        // if no spine item specified we are talking about current spine item
        if(pageRequest.spineItem && pageRequest.spineItem != _currentSpineItem) {
            if (!_currentSpineItem || !_currentSpineItem.href || _currentSpineItem.href != pageRequest.spineItem.href) {
                _deferredPageRequest = pageRequest;
                loadSpineItem(pageRequest.spineItem);
                return;
            }
        }
        
        _paginationInfo.columnCount = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) + _paginationInfo.columnGap) / (_paginationInfo.columnWidth + _paginationInfo.columnGap);
        _paginationInfo.columnCount = Math.round(_paginationInfo.columnCount);

        var totalGaps = (_paginationInfo.columnCount-1) * _paginationInfo.columnGap;
        var colWidthCheck = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) - totalGaps) / _paginationInfo.columnCount;
        colWidthCheck = Math.round(colWidthCheck);

        if (colWidthCheck > _paginationInfo.columnWidth)
        {
            console.debug("ADJUST COLUMN");
            console.log(_paginationInfo.columnWidth);
            console.log(colWidthCheck);
            
            _paginationInfo.columnWidth = colWidthCheck;
        }

        _paginationInfo.spreadCount =  Math.ceil(_paginationInfo.columnCount / _paginationInfo.visibleColumnCount);

        var pageIndex = undefined;

        if(pageRequest.spineItemPageIndex !== undefined) {
            pageIndex = pageRequest.spineItemPageIndex;
        }
        else if(pageRequest.elementId) {
            pageIndex = _navigationLogic.getPageForElementId(pageRequest.elementId);
        }
        else if(pageRequest.elementCfi) {
            if (pageRequest.elementCfi.indexOf("@") != -1) {
                try
                {
                    pageIndex = _navigationLogic.getPageForElementCfi(pageRequest.elementCfi,
                        ["cfi-marker", "mo-cfi-highlight"],
                        [],
                        ["MathJax_Message"]);
                }
                catch (e)
                {
                    pageIndex = 0;
                    console.error(e);
                }
            }
            else {
                try
                {
                    pageIndex = _navigationLogic.getPageForRangeCfi(pageRequest.elementCfi,
                        ["cfi-marker", "mo-cfi-highlight"],
                        [],
                        ["MathJax_Message"]);
                }
                catch (e)
                {
                    pageIndex = 0;
                    console.error(e);
                }
            }
        }
        else if(pageRequest.firstPage) {
            pageIndex = 0;
        }
        else if(pageRequest.lastPage) {
            pageIndex = _paginationInfo.columnCount - 1;
        }
        else {
            console.debug("No criteria in pageRequest");
            pageIndex = 0;
        }
        
        if (pageIndex < 0) {
            pageIndex = 0;
        }
        else if (pageIndex >= _paginationInfo.columnCount) {
            pageIndex = _paginationInfo.columnCount - 1;
        }

        if(pageIndex >= 0 && pageIndex < _paginationInfo.columnCount) {
            _paginationInfo.currentSpreadIndex = Math.floor(pageIndex / _paginationInfo.visibleColumnCount) ;
            onPaginationChanged(pageRequest.initiator, pageRequest.spineItem, pageRequest.elementId);
        }
        else {
            console.log('Illegal pageIndex value: '+ pageIndex + 'column count is ' + _paginationInfo.columnCount);
        }
    };

    function redraw() {

        var offsetVal =  -_paginationInfo.pageOffset + "px";

        if (_htmlBodyIsVerticalWritingMode)
        {
            _$epubHtml.css("top", offsetVal);
        }
        else
        {
            var ltr = _htmlBodyIsLTRDirection || _htmlBodyIsLTRWritingMode;
            
            _$epubHtml.css("left", ltr ? offsetVal : "");
            _$epubHtml.css("right", !ltr ? offsetVal : "");
        }

        showBook(); // as it's no longer hidden by shifting the position
    }

    function updateViewportSize() {

        var newWidth = _$contentFrame.width();
        var newHeight = _$contentFrame.height();

        if(_lastViewPortSize.width !== newWidth || _lastViewPortSize.height !== newHeight){

            _lastViewPortSize.width = newWidth;
            _lastViewPortSize.height = newHeight;
            return true;
        }

        return false;
    }

    function onPaginationChanged(initiator, paginationRequest_spineItem, paginationRequest_elementId) {

        _paginationInfo.pageOffset = (_paginationInfo.columnWidth + _paginationInfo.columnGap) * _paginationInfo.visibleColumnCount * _paginationInfo.currentSpreadIndex;
        
        redraw();
        self.trigger(ReadiumSDK.InternalEvents.CURRENT_VIEW_PAGINATION_CHANGED, { paginationInfo: self.getPaginationInfo(), initiator: initiator, spineItem: paginationRequest_spineItem, elementId: paginationRequest_elementId } );
    }

    this.openPagePrev = function (initiator) {

        if(!_currentSpineItem) {
            return;
        }
        
        _paginationInfo.columnCount = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) + _paginationInfo.columnGap) / (_paginationInfo.columnWidth + _paginationInfo.columnGap);
        _paginationInfo.columnCount = Math.round(_paginationInfo.columnCount);

        var totalGaps = (_paginationInfo.columnCount-1) * _paginationInfo.columnGap;
        var colWidthCheck = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) - totalGaps) / _paginationInfo.columnCount;
        colWidthCheck = Math.round(colWidthCheck);

        if (colWidthCheck > _paginationInfo.columnWidth)
        {
            console.debug("ADJUST COLUMN");
            console.log(_paginationInfo.columnWidth);
            console.log(colWidthCheck);
            
            _paginationInfo.columnWidth = colWidthCheck;
        }

        _paginationInfo.spreadCount =  Math.ceil(_paginationInfo.columnCount / _paginationInfo.visibleColumnCount);

        if(_paginationInfo.currentSpreadIndex > 0) {
            _paginationInfo.currentSpreadIndex--;
            onPaginationChanged(initiator);
        }
        else {

            var prevSpineItem = _spine.prevItem(_currentSpineItem, true);
            if(prevSpineItem) {

                var pageRequest = new ReadiumSDK.Models.PageOpenRequest(prevSpineItem, initiator);
                pageRequest.setLastPage();
                self.openPage(pageRequest);
            }
        }
    };

    this.openPageNext = function (initiator) {

        if(!_currentSpineItem) {
            return;
        }
        
        _paginationInfo.columnCount = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) + _paginationInfo.columnGap) / (_paginationInfo.columnWidth + _paginationInfo.columnGap);
        _paginationInfo.columnCount = Math.round(_paginationInfo.columnCount);

        var totalGaps = (_paginationInfo.columnCount-1) * _paginationInfo.columnGap;
        var colWidthCheck = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) - totalGaps) / _paginationInfo.columnCount;
        colWidthCheck = Math.round(colWidthCheck);

        if (colWidthCheck > _paginationInfo.columnWidth)
        {
            console.debug("ADJUST COLUMN");
            console.log(_paginationInfo.columnWidth);
            console.log(colWidthCheck);
            
            _paginationInfo.columnWidth = colWidthCheck;
        }

        _paginationInfo.spreadCount =  Math.ceil(_paginationInfo.columnCount / _paginationInfo.visibleColumnCount);

        if(_paginationInfo.currentSpreadIndex < _paginationInfo.spreadCount - 1) {
            _paginationInfo.currentSpreadIndex++;
            onPaginationChanged(initiator);
        }
        else {

            var nextSpineItem = _spine.nextItem(_currentSpineItem, true);
            if(nextSpineItem) {

                var pageRequest = new ReadiumSDK.Models.PageOpenRequest(nextSpineItem, initiator);
                pageRequest.setFirstPage();
                self.openPage(pageRequest);
            }
        }
    };
    
   


    function updatePagination() {

        var isDoublePageSyntheticSpread = ReadiumSDK.Helpers.deduceSyntheticSpread(_$viewport, _currentSpineItem, _viewSettings);
        
        var forced = (isDoublePageSyntheticSpread === false) || (isDoublePageSyntheticSpread === true);
        // excludes 0 and 1 truthy values which denote non-forced result
        
// console.debug("isDoublePageSyntheticSpread: " + isDoublePageSyntheticSpread);
// console.debug("forced: " + forced);
//
        if (isDoublePageSyntheticSpread === 0)
        {
            isDoublePageSyntheticSpread = 1; // try double page, will shrink if doesn't fit
// console.debug("TRYING SPREAD INSTEAD OF SINGLE...");
        }
        
        _paginationInfo.visibleColumnCount = _htmlBodyIsVerticalWritingMode ? 1 : (isDoublePageSyntheticSpread ? 2 : 1);
   
        if(!_$epubHtml) {
            return;
        }
        
        hideBook(); // shiftBookOfScreen();
        
        var borderLeft = parseInt(_$viewport.css("border-left-width"));
        var borderRight = parseInt(_$viewport.css("border-right-width"));
        var adjustedGapLeft = _paginationInfo.columnGap/2;
        adjustedGapLeft = Math.max(0, adjustedGapLeft-borderLeft)
        var adjustedGapRight = _paginationInfo.columnGap/2;
        adjustedGapRight = Math.max(0, adjustedGapRight-borderRight)

        var filler = 0;
        
        // At 100% font-size = 16px (on HTML, not body or descendant markup!)
        var MAXW = 550; //TODO user/vendor-configurable?
        var MINW = 400;
        
//         var win = _$iframe[0].contentDocument.defaultView || _$iframe[0].contentWindow;
//         var htmlBodyComputedStyle = win.getComputedStyle(_$htmlBody[0], null);
//         if (htmlBodyComputedStyle)
//         {
//             var fontSize = undefined;
//             if (htmlBodyComputedStyle.getPropertyValue)
//             {
//                 fontSize = htmlBodyComputedStyle.getPropertyValue("font-size");
//             }
//             else
//             {
//                 fontSize = htmlBodyComputedStyle.fontSize;
//             }
// console.debug(fontSize);
//         }
        /*
        if (_viewSettings.fontSize)
        {
            var fontSizeAdjust = (_viewSettings.fontSize*0.8)/100;
            MAXW = Math.floor(MAXW * fontSizeAdjust);
            MINW = Math.floor(MINW * fontSizeAdjust);
        }
        */
        var availableWidth = _$viewport.width();
        var textWidth = availableWidth - borderLeft - borderRight - adjustedGapLeft - adjustedGapRight;
//        console.debug("availableWidth=" + availableWidth + ",textWidth=" + textWidth);
        if (isDoublePageSyntheticSpread)
        {
            textWidth = (textWidth - _paginationInfo.columnGap) * 0.5;
        }
        
        if (textWidth > MAXW * ratio)
        {
 console.debug("LIMITING WIDTH");
            filler = Math.floor((textWidth - MAXW * ratio) * (isDoublePageSyntheticSpread ? 1 : 0.5));
        }
        else if (!forced && textWidth < MINW * ratio && isDoublePageSyntheticSpread)
        {
console.debug("REDUCING SPREAD TO SINGLE");
            isDoublePageSyntheticSpread = false;
            _paginationInfo.visibleColumnCount = 1;
            
            textWidth = availableWidth - borderLeft - borderRight - adjustedGapLeft - adjustedGapRight;
            if (textWidth > MAXW * ratio)
            {
                filler = Math.floor((textWidth - MAXW * ratio) * 0.5);
            }
        }
        
        _$el.css({"left": (filler+adjustedGapLeft + "px"), "right": (filler+adjustedGapRight + "px")});
        updateViewportSize(); //_$contentFrame ==> _lastViewPortSize
        
        _$iframe.css("width", _lastViewPortSize.width + "px");
        _$iframe.css("height", _lastViewPortSize.height + "px");

        _$epubHtml.css("height", _lastViewPortSize.height + "px");

        //normalise spacing to avoid interference with column-isation
        _$epubHtml.css('margin', 0);
        _$epubHtml.css('padding', 0);
        _$epubHtml.css('border', 0);
        _$htmlBody.css('margin', 0);
        _$htmlBody.css('padding', 0);

/* comment the following code which cause wrong pagination
        var spacing = 0;
        try
        {
            spacing = parseInt(_$htmlBody.css('padding-top')) + parseInt(_$htmlBody.css('border-top-width')) + parseInt(_$htmlBody.css('border-bottom-width'));
        }
        catch(err)
        {
            
        }
        // Needed for Firefox, otherwise content shrinks vertically, resulting in scrollWidth accomodating more columns than necessary
        //_$htmlBody.css("min-height", _lastViewPortSize.height-spacing-9 + "px");
        _$htmlBody.css("min-height", "50%");
        _$htmlBody.css("max-height", _lastViewPortSize.height-spacing + "px");
*/

        _paginationInfo.rightToLeft = _spine.isRightToLeft();

        _paginationInfo.columnWidth = Math.round(((_htmlBodyIsVerticalWritingMode ? _lastViewPortSize.height : _lastViewPortSize.width) - _paginationInfo.columnGap * (_paginationInfo.visibleColumnCount - 1)) / _paginationInfo.visibleColumnCount);
//        console.debug("columnWidth=" + _paginationInfo.columnWidth);

        _$epubHtml.css("width", (_htmlBodyIsVerticalWritingMode ? _lastViewPortSize.width : _paginationInfo.columnWidth) + "px");

        _.each(['-webkit-', '-moz-', '-ms-', ''], function(prefix) {
            _$epubHtml.css(prefix + "column-width", _paginationInfo.columnWidth + "px");
        });

        _$epubHtml.css({left: "0", right: "0", top: "0"});
        
        if (_origImageSize.length == 0) {
            getImageSize();
            adjustImageSize();
        }
        
        ReadiumSDK.Helpers.triggerLayout(_$iframe);

//        console.debug("scrollWidth=" + _$epubHtml[0].scrollWidth);
        _paginationInfo.columnCount = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) + _paginationInfo.columnGap) / (_paginationInfo.columnWidth + _paginationInfo.columnGap);
        _paginationInfo.columnCount = Math.round(_paginationInfo.columnCount);

        var totalGaps = (_paginationInfo.columnCount-1) * _paginationInfo.columnGap;
        var colWidthCheck = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) - totalGaps) / _paginationInfo.columnCount;
        colWidthCheck = Math.round(colWidthCheck);

        if (colWidthCheck > _paginationInfo.columnWidth)
        {
            console.debug("ADJUST COLUMN");
            console.log(_paginationInfo.columnWidth);
            console.log(colWidthCheck);
            
            _paginationInfo.columnWidth = colWidthCheck;
        }

        _paginationInfo.spreadCount =  Math.ceil(_paginationInfo.columnCount / _paginationInfo.visibleColumnCount);

        if(_paginationInfo.currentSpreadIndex >= _paginationInfo.spreadCount) {
            _paginationInfo.currentSpreadIndex = _paginationInfo.spreadCount - 1;
        }

        if(_deferredPageRequest) {

            //if there is a request for specific page we get here
            openDeferredElement();
        }
        else {

            //we get here on resizing the viewport
            
            onPaginationChanged(self); // => redraw() => showBook(), so the trick below is not needed

            // //We do this to force re-rendering of the document in the iframe.
            // //There is a bug in WebView control with right to left columns layout - after resizing the window html document
            // //is shifted in side the containing div. Hiding and showing the html element puts document in place.
            // _$epubHtml.hide();
            // setTimeout(function() {
            //     _$epubHtml.show();
            //     onPaginationChanged(self); // => redraw() => showBook()
            // }, 50);

        }
    }

//    function shiftBookOfScreen() {
//
//        if(_spine.isLeftToRight()) {
//            _$epubHtml.css("left", (_lastViewPortSize.width + 1000) + "px");
//        }
//        else {
//            _$epubHtml.css("right", (_lastViewPortSize.width + 1000) + "px");
//        }
//    }

    function hideBook()
    {
        if (_currentOpacity != -1) return; // already hidden
        
        _currentOpacity = _$epubHtml.css('opacity');
        _$epubHtml.css('opacity', "0");
    }

    function showBook()
    {
        if (_currentOpacity != -1)
        {
            _$epubHtml.css('opacity', _currentOpacity);
        }
        _currentOpacity = -1;
    }

    this.getFirstVisibleElementCfi = function() {

        var contentOffsets = getVisibleContentOffsets();
        return _navigationLogic.getFirstVisibleElementCfi(contentOffsets);
    };

    this.getPaginationInfo = function() {
        var paginationInfo = new ReadiumSDK.Models.CurrentPagesInfo(_spine, false);

        if(!_currentSpineItem) {
            return paginationInfo;
        }

        var pageIndexes = getOpenPageIndexes();

        for(var i = 0, count = pageIndexes.length; i < count; i++) {

            paginationInfo.addOpenPage(pageIndexes[i], _paginationInfo.columnCount, _currentSpineItem.idref, _currentSpineItem.index);
        }

        return paginationInfo;

    };

    function getOpenPageIndexes() {

        var indexes = [];

        var currentPage = _paginationInfo.currentSpreadIndex * _paginationInfo.visibleColumnCount;

        for(var i = 0; i < _paginationInfo.visibleColumnCount && (currentPage + i) < _paginationInfo.columnCount; i++) {

            indexes.push(currentPage + i);
        }

        return indexes;

    }
    
    function updateMedias() {

        if(!_$epubHtml) {
            return;
        }
        /*
        $('audio', _$epubHtml).each(function(){
            $elem = $(this);
            
            $elem.replaceWith('<img class=\'audio_to_img\' src=\'file:///android_asset/epub_audio_player.png\' data-orig=\'' + $elem.attr('src') + '\' />');
        });
        
        $('video', _$epubHtml).each(function(){
            $elem = $(this);
            
            $elem.replaceWith('<img class=\'video_to_img\' src=\'file:///android_asset/epub_video_player.png\' data-orig=\'' + $elem.attr('src') + '\' />');
        });
        */
        
        $('img.FragmentationVideo', _$epubHtml).each(function(){
            $elem = $(this);
            
            $elem.attr('src', 'file:///android_asset/epub_video_player.png');
        });
        
        $('img.FragmentationAudio', _$epubHtml).each(function(){
            $elem = $(this);
            
            $elem.attr('src', 'file:///android_asset/audio_player_0000.png');
        });
        
        $('img.FragmentationQuestion', _$epubHtml).each(function(){
            $elem = $(this);
            
            $elem.attr('src', 'file:///android_asset/epub_exercise.png');
        });
    }
    
    //we need this styles for css columnizer not to chop big images
    function resizeImages() {

        if(!_$epubHtml) {
            return;
        }
        
//        console.debug("resizeImage start");
        _origImageSize = [];
        var index = 0;

        $('img', _$epubHtml).each(function(){
        
            var $elem;
            
            $elem = $(this);
            
            if ($elem.is("img.FragmentationAudio") || $elem.is("img.FragmentationVideo")) {
                $elem.css('max-width', '100%');
                $elem.css('max-height', '100%');

                if(!$elem.css('height')) {
                    $elem.css('height', 'auto');
                }

                if(!$elem.css('width')) {
                    $elem.css('width', 'auto');
                }
            } else if ($elem.is("img.FragmentationQuestion")) {
                $elem.css('max-width', '50%');
                $elem.css("margin-left", '25%');

                if(!$elem.css('height')) {
                    $elem.css('height', 'auto');
                }

                if(!$elem.css('width')) {
                    $elem.css('width', 'auto');
                }
            } else {
            
            /*
            // if we set max-width/max-height to 100% columnizing engine chops images embedded in the text
            // (but not if we set it to 99-98%) go figure.
            // TODO: CSS min-w/h is content-box, not border-box (does not take into account padding + border)? => images may still overrun?
            $elem.css('max-width', '98%');
            $elem.css('max-height', '98%');

            if(!$elem.css('height')) {
                $elem.css('height', 'auto');
            }

            if(!$elem.css('width')) {
                $elem.css('width', 'auto');
            }
            */
            
            //_origImageSize[index++] = $elem.width(); 
            }
        });
        /*
        $('video', _$epubHtml).each(function(){

            var $elem;
            
            $elem = $(this);
            
            // if we set max-width/max-height to 100% columnizing engine chops images embedded in the text
            // (but not if we set it to 99-98%) go figure.
            // TODO: CSS min-w/h is content-box, not border-box (does not take into account padding + border)? => images may still overrun?
            $elem.css('max-width', '98%');
            $elem.css('max-height', '98%');

                $elem.css('height', 'auto');

            if(!$elem.css('width')) {
                $elem.css('width', 'auto');
            }
            
        });
        */
//        console.debug("resizeImage end");
    }
    
    function getImageSize() {
    
        if(!_$epubHtml) {
            return;
        }
        
        var index = 0;
        _origImageSize = [];

        $('img', _$epubHtml).each(function(){

            var $elem;
            
            $elem = $(this); 
            
            if ($elem.is("img.FragmentationAudio") || $elem.is("img.FragmentationVideo") || $elem.is("img.FragmentationQuestion")) {
            } else {
            var w = $elem.width();
            var h = $elem.height();
            var size = {};
            size.width = w;
            size.height = h;
            _origImageSize[index++] = size;
            }
        });
    }
    
    function adjustImageSize() {

        if(!_$epubHtml) {
            return;
        }

//        console.debug("adjustImageSize start");
        var index = 0;
        $('img', _$epubHtml).each(function(){
            var $elem;
            $elem = $(this);
            
            if ($elem.is("img.FragmentationAudio") || $elem.is("img.FragmentationVideo") || $elem.is("img.FragmentationQuestion")) {
            } else {
            
            if (_origImageSize[index]) {

                var newWidth  = _origImageSize[index].width*_fontSize*ratio/100;
                if (newWidth > _paginationInfo.columnWidth) {
                    newWidth = _paginationInfo.columnWidth;
                }
                var newHeight = newWidth * _origImageSize[index].height / _origImageSize[index].width;
                if (newHeight > _lastViewPortSize.height) {
                    newHeight = _lastViewPortSize.height;
                    newWidth = newHeight * _origImageSize[index].width / _origImageSize[index].height; 
                }
                $elem.width(newWidth);
                //$elem.height(newHeight);
                $elem.css('height', 'auto');
            }
            index++;
            
            }
        });
//        console.debug("adjustImageSize end");
    }

    this.bookmarkCurrentPage = function() {

        if(!_currentSpineItem) {

            return new ReadiumSDK.Models.BookmarkData("", "", "", 0);
        }
        
        var epubContentDocument = _$iframe[0].contentDocument;
        var range = epubContentDocument.caretRangeFromPoint(0, 0);
        range.expand("sentence");
        var text = range.toString();

        return new ReadiumSDK.Models.BookmarkData(_currentSpineItem.idref, self.getFirstVisibleElementCfi(), text,
                function getTimeStamp() {
                    var now = new Date();
                    return now.getTime();
                }());
    };
    
    function getVisibleContentOffsets() {
        //TODO: _htmlBodyIsVerticalWritingMode ? (_lastViewPortSize.height * _paginationInfo.currentSpreadIndex)
        // NOT used with options.rectangleBased anyway (see CfiNavigationLogic constructor call, here in this reflow engine class)
        var columnsLeftOfViewport = Math.round(_paginationInfo.pageOffset / (_paginationInfo.columnWidth + _paginationInfo.columnGap));

        var topOffset =  columnsLeftOfViewport * _$contentFrame.height();
        var bottomOffset = topOffset + _paginationInfo.visibleColumnCount * _$contentFrame.height();

        return {top: topOffset, bottom: bottomOffset};
    }

    this.getLoadedSpineItems = function() {
        return [_currentSpineItem];
    };

    this.getElementByCfi = function(spineItem, cfi, classBlacklist, elementBlacklist, idBlacklist) {

        if(spineItem != _currentSpineItem) {
            console.error("spine item is not loaded");
            return undefined;
        }

        return _navigationLogic.getElementByCfi(cfi, classBlacklist, elementBlacklist, idBlacklist);
    };

    this.getElementById = function(spineItem, id) {

        if(spineItem != _currentSpineItem) {
            console.error("spine item is not loaded");
            return undefined;
        }

        return _navigationLogic.getElementById(id);
    };

    this.getElement = function(spineItem, selector) {

        if(spineItem != _currentSpineItem) {
            console.error("spine item is not loaded");
            return undefined;
        }

        return _navigationLogic.getElement(selector);
    };

    this.getFirstVisibleMediaOverlayElement = function() {

        var visibleContentOffsets = getVisibleContentOffsets();
        return _navigationLogic.getFirstVisibleMediaOverlayElement(visibleContentOffsets);
    };
    
    // /**
    //  * @deprecated
    //  */
    // this.getVisibleMediaOverlayElements = function() {
    // 
    //     var visibleContentOffsets = getVisibleContentOffsets();
    //     return _navigationLogic.getVisibleMediaOverlayElements(visibleContentOffsets);
    // };

    this.insureElementVisibility = function(spineItemId, element, initiator) {

        var $element = $(element);
        if(_navigationLogic.isElementVisible($element, getVisibleContentOffsets()))
        {
            return;
        }

        var page = _navigationLogic.getPageForElement($element);

        if(page == -1)
        {
            return;
        }

        var openPageRequest = new ReadiumSDK.Models.PageOpenRequest(_currentSpineItem, initiator);
        openPageRequest.setPageIndex(page);

        var id = element.id;
        if (!id)
        {
            id = element.getAttribute("id");
        }
        
        if (id)
        {
            openPageRequest.setElementId(id);
        }

        self.openPage(openPageRequest);
    }

    this.updateSelection = function(x1, y1, x2, y2, dragging, timestamp) {
        console.debug("updateSelection x1=" + x1 + ",y1=" + y1 + ",x2=" + x2 + ",y2=" + y2 + ",dragging=" + dragging + ",timestamp=" + timestamp);
        var padding = $(document.body).padding();
        var gap = _$el.position();
        
        if (x1 < 0) {
            x1 = 0;
        }
        else if (x1 > _$contentFrame.width() - 1) {
            x1 = _$contentFrame.width() - 1;
        }
        if (y1 < 0) {
            y1 = 0;
        }
        else if (y1 > _$contentFrame.height() - 1) {
            y1 =  _$contentFrame.height() - 1;
        }
        
        if (x2 < 0) {
            x2 = 0;
        }
        else if (x2 > _$contentFrame.width() - 1) {
            x2 = _$contentFrame.width() - 1;
        }
        if (y2 < 0) {
            y2 = 0;
        }
        else if (y2 > _$contentFrame.height() - 1) {
            y2 =  _$contentFrame.height() - 1;
        }
        
        var epubContentDocument = _$iframe[0].contentDocument;
        var range1 = rangy.createRangeFromPoints(x1, y1, x1, y1, epubContentDocument);
        var range2 = rangy.createRangeFromPoints(x2, y2, x2, y2, epubContentDocument);
        var range;
        
        if (range1.compareBoundaryPoints(Range.START_TO_END, range2) > 0) {
            var range = rangy.createRangeFromPoints(x2, y2, x1, y1, epubContentDocument);
        }
        else {
            var range = rangy.createRangeFromPoints(x1, y1, x2, y2, epubContentDocument);
        }
        
//        var sel = rangy.getSelection(epubContentDocument);

        if (dragging && _selectionRange && (range.startContainer == _selectionRange.startContainer) && (range.startOffset == _selectionRange.startOffset)
                && (range.endContainer == _selectionRange.endContainer) && (range.endOffset == _selectionRange.endOffset))
        {
            return;
        }

        if (range.startContainer.nodeType != 3 || range.endContainer.nodeType != 3 || !isRangeFullyVisible(range)) {
            if (dragging) {
                return;
            }
        }
        else {
            /*
            var startNode = range.startContainer;
            var startOffset = range.startOffset;
            var endNode = range.endContainer;
            var endOffset = range.endOffset;

            var pageRange = rangy.createRangeFromPoints(0, 0, _$contentFrame.width() - 1, _$contentFrame.height() - 1, epubContentDocument);
            if (startNode == pageRange.startContainer && startOffset < pageRange.startOffset) {
                startOffset = pageRange.startOffset;
                range.setStart(startNode, startOffset);
            }
            if (endNode == pageRange.endContainer && endOffset > pageRange.endOffset) {
                endOffset = pageRange.endOffset;
                range.setEnd(endNode, endOffset);
            }
            */

            var text = range.toString();
            console.debug("text=" + text);
            if (text.length == 0) {
                if (dragging) {
                    return;
                }
            } else {
//                var handleBounds = {};
//                var startPos = range.getStartClientPos();
//                var endPos = range.getEndClientPos();
//                if (startPos.left >= _$contentFrame.width() || startPos.right < 0 || endPos.left >= _$contentFrame.width() || endPos.right < 0) {
//                    return;
//                if (startPos.x < 0) {
//                    handleBounds.left = padding.left + gap.left;
//                    handleBounds.top = padding.top + gap.top;
//                } else {
//                    handleBounds.left = startPos.x + padding.left + gap.left;
//                    handleBounds.top = startPos.y + padding.top + gap.top;
//                }
//                if (endPos.x >= _$contentFrame.width()) {
//                    handleBounds.right = _$contentFrame.width() + padding.left + gap.left;
//                    handleBounds.bottom = _$contentFrame.height() + padding.top + gap.top;
//                } else {
//                    handleBounds.right = endPos.x + padding.left + gap.left;
//                    handleBounds.bottom = endPos.y + padding.top + gap.top;
//                }
//                console.debug("handleBounds=" + JSON.stringify(handleBounds));
//                sel.removeAllRanges();
//                sel.addRange(range);
                _selectionRange = range;
            }
        }
        if (!dragging) {
            try {
//            var cfi = ReadiumSDK.reader.getCurrentSelectionCfi();
//            var result = _navigationLogic.getRangeTargetElementsByCfi(cfi.cfi,
//                    ["cfi-marker", "mo-cfi-highlight"],
//                    [],
//                    ["MathJax_Message"]);
//            var check = (range.startContainer == result.startElement) && (range.startOffset == result.startOffset)
//                    && (range.endContainer == result.endElement) && (range.endOffset == result.endOffset);
//            console.debug("idref=" + cfi.idref + ", cfi=" + cfi.cfi + ", check=" + check);
            
            //var clients = getRangeVisibleClientRects(range);
//            sel.removeAllRanges();

            var startNode = _selectionRange.startContainer;
            var startOffset = _selectionRange.startOffset;
            var endNode = _selectionRange.endContainer;
            var endOffset = _selectionRange.endOffset;
            var cfi = EPUBcfi.Generator.generateCharOffsetRangeComponent(
                    startNode,
                    startOffset,
                    endNode,
                    endOffset,
                    ["cfi-marker", "mo-cfi-highlight"],
                    [],
                    ["MathJax_Message"]
                    );

            var text = _selectionRange.toString();
            var rect = getRangeVisibleBoundingClientRect(range);

            var note = new ReadiumSDK.Models.NoteData(_currentSpineItem.idref, cfi, text, rect, timestamp);
            self.trigger(ReadiumSDK.Events.GET_NOTE_DATA, note);
            
            var highlight = {};
            highlight.idref = note.idref;
            /*
            var elem = _navigationLogic.getRangeTargetElementsByCfi(note.contentCFI,
                    ["cfi-marker", "mo-cfi-highlight"],
                    [],
                    ["MathJax_Message"]);
            var range = rangy.createRange(epubContentDocument);
            range.setStart(elem.startElement, elem.startOffset);
            range.setEnd(elem.endElement, elem.endOffset);
            */
            highlight.range = _selectionRange;
            highlight.timestamp = note.timestamp;
            _highlights.push(highlight);
            }
            catch (e) {
//                sel.removeAllRanges();
                console.error(e);
            }
            
        } else {
            try {
//            var cfi = ReadiumSDK.reader.getCurrentSelectionCfi();
//            var result = _navigationLogic.getRangeTargetElementsByCfi(cfi.cfi,
//                    ["cfi-marker", "mo-cfi-highlight"],
//                    [],
//                    ["MathJax_Message"]);
//            var check = (range.startContainer == result.startElement) && (range.startOffset == result.startOffset)
//                    && (range.endContainer == result.endElement) && (range.endOffset == result.endOffset);
//            console.debug("idref=" + cfi.idref + ", cfi=" + cfi.cfi + ", check=" + check);

            var clients = getRangeVisibleClientRects(_selectionRange);
//            sel.removeAllRanges();

            self.trigger(ReadiumSDK.Events.DRAW_SELECTION, clients);
            }
            catch (e) {
//                sel.removeAllRanges();
                console.error(e);
            }
        }

    }
    
    function getRangeVisibleClientRects(range) {
        if (range.nativeRange)
            range = range.nativeRange;
        var clientRects = range.getClientRects();
        var clients = [];
        var padding = $(document.body).padding();
        var gap = _$el.position();
        for (var i = 0; i < clientRects.length; i++) {
            var rect = clientRects[i];
            var next = clientRects[i + 1];

            if (next && (rect.top + rect.height >= next.top + next.height) && (rect.left + rect.width >= next.left + next.width))
                continue;

            if (rect.left >= _$contentFrame.width() || rect.right < 0) {
                continue;
            }

            if (rect.top >= _$contentFrame.height() || rect.bottom < 0) {
                continue;
            }

            if (rect.left < 0 || rect.top < 0) {
                continue;
            }

            if (rect.width == 0 || rect.height == 0) {
                continue;
            }

            var client = {};
            client.left = rect.left + padding.left + gap.left;
            client.top = rect.top + padding.top + gap.top;
            client.width = rect.width;
            client.height = rect.height;

            clients.push(client);
        }
        return clients;
    }

    function getRangeVisibleBoundingClientRect(range) {
            if (range.nativeRange)
                range = range.nativeRange;
            var clientRects = range.getClientRects();
            var padding = $(document.body).padding();
            var gap = _$el.position();
            var rectLeft = null;
            var rectTop = null;
            var rectRight = null;
            var rectBottom = null;
            var result = {};
            for (var i = 0; i < clientRects.length; i++) {
                var rect = clientRects[i];
                var next = clientRects[i + 1];

                if (next && (rect.top + rect.height >= next.top + next.height) && (rect.left + rect.width >= next.left + next.width))
                    continue;

                if (rect.left >= _$contentFrame.width() || rect.right < 0) {
                    continue;
                }

                if (rect.top >= _$contentFrame.height() || rect.bottom < 0) {
                    continue;
                }

                if (rect.left < 0 || rect.top < 0) {
                    continue;
                }

                if (rect.width == 0 || rect.height == 0) {
                    continue;
                }

                var left = rect.left + padding.left + gap.left;
                var top = rect.top + padding.top + gap.top;
                var right = rect.right + padding.left + gap.left;
                var bottom = rect.bottom + padding.top + gap.top;

                if (!rectLeft || rectLeft > left)
                    rectLeft = left;
                if (!rectTop || rectTop > top)
                    rectTop = top;
                if (!rectRight || rectRight < right)
                    rectRight = right;
                if (!rectBottom || rectBottom < bottom)
                    rectBottom = bottom;
            }

            result.left = rectLeft ? rectLeft : padding.left + gap.left;
            result.top = rectTop ? rectTop : padding.top + gap.top;
            result.width = rectLeft && rectRight != null ? rectRight - rectLeft : 0;
            result.height = rectTop && rectBottom != null ? rectBottom - rectTop : 0;
            return result;
        }

    function isRangeFullyVisible(range) {
        if (range.nativeRange)
            range = range.nativeRange;
        var clientRects = range.getClientRects();
        for (var i = 0; i < clientRects.length; i++) {
            var rect = clientRects[i];

            if (rect.top < 0 && rect.bottom > 0)
                continue;

            if (rect.left >= _$contentFrame.width() || rect.right < 0) {
                return false;
            }

            if (rect.top >= _$contentFrame.height() || rect.bottom < 0) {
                return false;
            }

            if (rect.left < 0 || rect.top < 0) {
                return false;
            }
        }
        return true;
    }
    
    this.updateHighlights = function(highlights) {
//        console.debug("updateHighlights here");
        var notes = JSON.parse(highlights);
//        console.debug("updateHighlights after path");
        _highlights = [];
        var epubContentDocument = _$iframe[0].contentDocument;
        for (var i = 0; i < notes.length; i++) {
            var note = notes[i];
            if (note.idref == _currentSpineItem.idref) {
                var highlight = {};
                highlight.idref = note.idref;
                var elem = _navigationLogic.getRangeTargetElementsByCfi(note.contentCFI,
                        ["cfi-marker", "mo-cfi-highlight"],
                        [],
                        ["MathJax_Message"]);
                var range = rangy.createRange(epubContentDocument);
                range.setStart(elem.startElement, elem.startOffset);
                range.setEnd(elem.endElement, elem.endOffset);
                highlight.range = range;
                highlight.timestamp = note.timestamp;
                _highlights.push(highlight);
            }
        }
    }
    
    this.drawHighlights = function() {
        var results = [];
        for (var i = 0; i < _highlights.length; i++) {
            var result = {};
            var highlight = _highlights[i];
            result.timestamp = highlight.timestamp;
            if (highlight.idref == _currentSpineItem.idref) {
                var rects = getRangeVisibleClientRects(highlight.range);
                result.rects = rects;
                results.push(result);
            }
        }
        self.trigger(ReadiumSDK.Events.DRAW_HIGHLIGHTS, results);
    }
    
    this.updateBookmarks = function(bookmarks) {
//        console.debug("updateBookmarks here");
        var arr = JSON.parse(bookmarks);
//        console.debug("updateBookmarks after path");
        _bookmarks = [];
        var epubContentDocument = _$iframe[0].contentDocument;
        for (var i = 0; i < arr.length; i++) {
            var bookmark = arr[i];
            if (bookmark.idref == _currentSpineItem.idref) {
                var item = {};
                item.timestamp = bookmark.timestamp;
                item.idref = bookmark.idref;
                var elem = _navigationLogic.getElementAndOffsetByCfi(bookmark.contentCFI,
                        ["cfi-marker", "mo-cfi-highlight"],
                        [],
                        ["MathJax_Message"]);
                item.x = elem.x;
                item.y = elem.y;
                item.$element = elem.$element;
                _bookmarks.push(item);
            }
        }
    }
    
    this.checkBookmarks = function() {
        var result = isCurrentPageBookmarked();
        self.trigger(ReadiumSDK.Events.CHECK_BOOKMARKS, result);
    }
    
    this.removeBookmarks = function() {
        var results = [];
        for (var i = 0; i < _bookmarks.length; i++) {
            var bookmark = _bookmarks[i];
            if (bookmark.idref == _currentSpineItem.idref) {
                var page = _navigationLogic.getPageForPointOnElement(bookmark.$element, bookmark.x, bookmark.y);
                if (page == _paginationInfo.currentSpreadIndex) {
                    var item = {};
                    item.timestamp = bookmark.timestamp;
                    results.push(item);
                }
            }
        }
        self.trigger(ReadiumSDK.Events.REMOVE_BOOKMARKS, results);
    }
    
    function isCurrentPageBookmarked() {
        var result = false;
        for (var i = 0; i < _bookmarks.length; i++) {
            var bookmark = _bookmarks[i];
            if (bookmark.idref == _currentSpineItem.idref) {
                var page = _navigationLogic.getPageForPointOnElement(bookmark.$element, bookmark.x, bookmark.y);
                if (page == _paginationInfo.currentSpreadIndex) {
                    result = true;
                    break;
                }
            }
        }
        return result;
    }

    this.singleTap = function(x, y) {
        console.debug("singleTap " + x + "," + y);

        var padding = $(document.body).padding();
        var gap = _$el.position();
        var realX = x - padding.left - gap.left;
        var realY = y - padding.top - gap.top;

        var epubContentDocument = _$iframe[0].contentDocument;
        var target = epubContentDocument.elementFromPoint(realX, realY);

        if ($(target).is("img.FragmentationAudio"))
        {
            var $elem = $(target);
            var pos = $elem.position();
            self.trigger(ReadiumSDK.Events.AUDIO_CLICKED, { id: $elem.data("id"), name: $elem.data("name"), left: (pos.left + padding.left + gap.left), top: (pos.top + padding.top + gap.top), width: $elem.width(), height: $elem.height(), doubleTap: false} );
        }
        else if ($(target).is("img.FragmentationVideo"))
        {
            var $elem = $(target);
            var pos = $elem.position();
            self.trigger(ReadiumSDK.Events.VIDEO_CLICKED, { id: $elem.data("id"), name: $elem.data("name"), left: (pos.left + padding.left + gap.left), top: (pos.top + padding.top + gap.top), width: $elem.width(), height: $elem.height(), doubleTap: false} );
        }
        else if ($(target).is("img.FragmentationQuestion"))
        {
            var $elem = $(target);
            self.trigger(ReadiumSDK.Events.EXERCISE_CLICKED, { knowledge: $elem.data("category"), exclude: $elem.data("question"), count: $elem.data("count"), doubleTap: false } );
        }
        else
        {
            var range = null;
            var intersection = null;
            if (realX >= 0 && realX < _$contentFrame.width() && realY >= 0 && realY < _$contentFrame.height()) {
                range = rangy.createRangeFromPoints(realX, realY, realX, realY, epubContentDocument);
                if (range) {
                    for (var i = 0; i < _highlights.length; i++) {
                        var highlight = _highlights[i];
                        if (range.intersectsOrTouchesRange(highlight.range)) {
                            intersection = highlight;
                            break;
                        }
                    }
                }
            }

            if (intersection) {
                //TODO highlight clicked

                var padding = $(document.body).padding();
                var gap = _$el.position();
                var pos = getRangeVisibleBoundingClientRect(intersection.range);
                self.trigger(ReadiumSDK.Events.HIGHLIGHT_CLICKED, { timestamp: intersection.timestamp, left: pos.left, top: pos.top, width: pos.width, height: pos.height} );
            }
            else {
                 _paginationInfo.columnCount = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) + _paginationInfo.columnGap) / (_paginationInfo.columnWidth + _paginationInfo.columnGap);
                 _paginationInfo.columnCount = Math.round(_paginationInfo.columnCount);

                 var totalGaps = (_paginationInfo.columnCount-1) * _paginationInfo.columnGap;
                 var colWidthCheck = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) - totalGaps) / _paginationInfo.columnCount;
                 colWidthCheck = Math.round(colWidthCheck);

                 if (colWidthCheck > _paginationInfo.columnWidth)
                 {
                     console.debug("ADJUST COLUMN");
                     console.log(_paginationInfo.columnWidth);
                     console.log(colWidthCheck);

                     _paginationInfo.columnWidth = colWidthCheck;
                 }

                 _paginationInfo.spreadCount =  Math.ceil(_paginationInfo.columnCount / _paginationInfo.visibleColumnCount);
                 var paginationInfo = self.getPaginationInfo();

                 if (realX < _$contentFrame.width() / 3) {
                    if (paginationInfo.canGoPrev()) {
                        self.trigger(ReadiumSDK.Events.TAP_LEFT);
                    } else {
                        self.trigger(ReadiumSDK.Events.TAP_LEFT_FIRST)
                    }
                }
                else if (realX > _$contentFrame.width() * 2 / 3) {
                    if (paginationInfo.canGoNext()) {
                        self.trigger(ReadiumSDK.Events.TAP_RIGHT);
                    } else {
                        self.trigger(ReadiumSDK.Events.TAP_RIGHT_LAST)
                    }
                }
                else {
                    self.trigger(ReadiumSDK.Events.TAP_CENTER);
                }
            }
        }
    }

    this.doubleTap = function(x, y) {
        console.debug("doubleTap " + x + "," + y);

        var padding = $(document.body).padding();
        var gap = _$el.position();
        var realX = x - padding.left - gap.left;
        var realY = y - padding.top - gap.top;

        var epubContentDocument = _$iframe[0].contentDocument;
        var target = epubContentDocument.elementFromPoint(realX, realY);

        if ($(target).is("img.FragmentationAudio"))
        {
            var $elem = $(target);
            var pos = $elem.position();
            self.trigger(ReadiumSDK.Events.AUDIO_CLICKED, { id: $elem.data("id"), name: $elem.data("name"), left: (pos.left + padding.left + gap.left), top: (pos.top + padding.top + gap.top), width: $elem.width(), height: $elem.height(), doubleTap: true} );
        }
        else if ($(target).is("img.FragmentationVideo"))
        {
            var $elem = $(target);
            var pos = $elem.position();
            self.trigger(ReadiumSDK.Events.VIDEO_CLICKED, { src: $elem.data("orig"), left: (pos.left + padding.left + gap.left), top: (pos.top + padding.top + gap.top), width: $elem.width(), height: $elem.height(), doubleTap: true} );
        }
        else if ($(target).is("img.FragmentationQuestion"))
        {
            var $elem = $(target);
            self.trigger(ReadiumSDK.Events.EXERCISE_CLICKED, { knowledge: $elem.data("category"), exclude: $elem.data("question"), count: $elem.data("count"), doubleTap: true } );
        }
        else if ($(target).is("img")) {
            var $elem = $(target);
            var pos = $elem.position();
            self.trigger(ReadiumSDK.Events.IMAGE_CLICKED, { src: $elem.attr("src"), left: (pos.left + padding.left + gap.left), top: (pos.top + padding.top + gap.top), width: $elem.width(), height: $elem.height(), doubleTap: true} );
        }
    }

    this.pinch = function(x, y) {
        console.debug("pinch " + x + "," + y);

        var padding = $(document.body).padding();
        var gap = _$el.position();
        var realX = x - padding.left - gap.left;
        var realY = y - padding.top - gap.top;

        var epubContentDocument = _$iframe[0].contentDocument;
        var target = epubContentDocument.elementFromPoint(realX, realY);

        var $elem = $(target);

        if ($elem.is("img.FragmentationAudio") || $elem.is("img.FragmentationVideo") || $elem.is("img.FragmentationQuestion")) {
        }
        else if ($elem.is("img"))
        {
            var pos = $elem.position();
            self.trigger(ReadiumSDK.Events.IMAGE_CLICKED, { src: $elem.attr("src"), left: (pos.left + padding.left + gap.left), top: (pos.top + padding.top + gap.top), width: $elem.width(), height: $elem.height(), doubleTap: false} );
        }
    }

    this.longPress = function(x, y) {
        console.debug("longPress " + x + "," + y);

        _selectionRange = null;

        var padding = $(document.body).padding();
        var gap = _$el.position();
        var realX = x - padding.left - gap.left;
        var realY = y - padding.top - gap.top;

        var epubContentDocument = _$iframe[0].contentDocument;

        if (realX < 0) {
            realX = 0;
        }
        else if (realX > _$contentFrame.width() - 1) {
            realX = _$contentFrame.width() - 1;
        }
        if (realY < 0) {
            realY = 0;
        }
        else if (realY > _$contentFrame.height() - 1) {
            realY =  _$contentFrame.height() - 1;
        }

        var range = rangy.createRangeFromPoints(realX, realY, realX, realY, epubContentDocument);
        if (range.startContainer.nodeType != 3 || range.endContainer.nodeType != 3 || !isRangeFullyVisible(range)) {
            return;
        }

        self.trigger(ReadiumSDK.Events.TAP_HOLD);
    }

    this.fling = function(direct) {
        console.debug("fling " + direct);

        _paginationInfo.columnCount = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) + _paginationInfo.columnGap) / (_paginationInfo.columnWidth + _paginationInfo.columnGap);
        _paginationInfo.columnCount = Math.round(_paginationInfo.columnCount);

        var totalGaps = (_paginationInfo.columnCount-1) * _paginationInfo.columnGap;
        var colWidthCheck = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) - totalGaps) / _paginationInfo.columnCount;
        colWidthCheck = Math.round(colWidthCheck);

        if (colWidthCheck > _paginationInfo.columnWidth)
        {
            console.debug("ADJUST COLUMN");
            console.log(_paginationInfo.columnWidth);
            console.log(colWidthCheck);

            _paginationInfo.columnWidth = colWidthCheck;
        }

        _paginationInfo.spreadCount =  Math.ceil(_paginationInfo.columnCount / _paginationInfo.visibleColumnCount);
        var paginationInfo = self.getPaginationInfo();
        if (direct == 1) {
            if (paginationInfo.canGoNext()) {
                self.trigger(ReadiumSDK.Events.TAP_RIGHT);
            } else {
                self.trigger(ReadiumSDK.Events.TAP_RIGHT_LAST)
            }
        } else if (direct == 2) {
            if (paginationInfo.canGoPrev()) {
                self.trigger(ReadiumSDK.Events.TAP_LEFT);
            } else {
                self.trigger(ReadiumSDK.Events.TAP_LEFT_FIRST)
            }
        }
    }

    this.updateSelection_raw = function(x1, y1, x2, y2, dragging, timestamp) {
        console.debug("updateSelection_raw x1=" + x1 + ",y1=" + y1 + ",x2=" + x2 + ",y2=" + y2 + ",dragging=" + dragging + ",timestamp=" + timestamp);

        var padding = $(document.body).padding();
        var gap = _$el.position();
        var realX1 = x1 - padding.left - gap.left;
        var realY1 = y1 - padding.top - gap.top;
        var realX2 = x2 - padding.left - gap.left;
        var realY2 = y2 - padding.top - gap.top;

        self.updateSelection(realX1, realY1, realX2, realY2, dragging, timestamp);
    }

    this.updatePaginationSimple = function() {
        console.debug("updatePaginationSimple");

        _paginationInfo.columnCount = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) + _paginationInfo.columnGap) / (_paginationInfo.columnWidth + _paginationInfo.columnGap);
        _paginationInfo.columnCount = Math.round(_paginationInfo.columnCount);

        var totalGaps = (_paginationInfo.columnCount-1) * _paginationInfo.columnGap;
        var colWidthCheck = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) - totalGaps) / _paginationInfo.columnCount;
        colWidthCheck = Math.round(colWidthCheck);

        if (colWidthCheck > _paginationInfo.columnWidth)
        {
            console.debug("ADJUST COLUMN");
            console.log(_paginationInfo.columnWidth);
            console.log(colWidthCheck);

            _paginationInfo.columnWidth = colWidthCheck;
        }

        _paginationInfo.spreadCount =  Math.ceil(_paginationInfo.columnCount / _paginationInfo.visibleColumnCount);
        var paginationInfo = self.getPaginationInfo();

        _paginationInfo.pageOffset = (_paginationInfo.columnWidth + _paginationInfo.columnGap) * _paginationInfo.visibleColumnCount * _paginationInfo.currentSpreadIndex;
        self.trigger(ReadiumSDK.Events.UPDATE_PAGINATION, { paginationInfo: self.getPaginationInfo() } );
    }

};
